//
//  lariat-online/include/TDCFragment.h    (W.Badgett)
//
//  Description of the  TDC raw data format
//
//
//  The event format is as such:
//  Per spill:  One ControllerHeader per spill
//              Up to 16 TdcSpillHeaders, as determined by ControllerHeader.tdcActiveMask 
//                  and ControllerHeader.nTDCs
//  Per event:  Number of events is ControllerHeader.triggerCounts
//              Each active TDC has one TdcEventHeader
//              After TdcEventHeader, there are TdcEventHeader.eventWordCount TdcHit records
//              The TdcHit records are considered part of the TdcEventHeader fragment
//              Bits <16:10> of the hit is the channel number
//              Bits <9:0> of thehit is the time bin
//

#ifndef _TDCFragment_h
#define _TDCFragment_h

#include <iostream>
#include <vector>

class TDCFragment
{
 public:

 enum
 {
   CHANNEL_NUMBER_S = 10,
   CHANNEL_NUMBER_M = 0x1F,
   TIME_BIN_S = 0,
   TIME_BIN_M = 0x3FF,
   TDC_EVENT_HEADER_SIZE = 9,
   MAX_HITS = 256,
   MAX_TDCS = 16,
   MAX_CHAMBERS = 4,
   N_CHANNELS = 64,
   TDCS_PER_CHAMBER = 4,
 };

 
#pragma pack(push,1)

  struct TdcControllerHeader
  {
    uint32_t fragmentSize;
    uint32_t fragmentType;
    uint32_t spillWordCount;
    uint32_t spillWordCountBuffer;
    uint32_t spillCounter;
    uint8_t controllerNumber;
    uint8_t  year;
    uint8_t  month;
    uint8_t  day;
    uint8_t  hours;
    uint8_t  minutes;
    uint8_t  seconds;
    uint32_t triggerCounts;
    uint32_t tdcStatus;
    uint32_t linkStatus;
    uint32_t tdcActiveMask;
    uint32_t nTDCs;
  } TdcControllerHeader_t;

  struct TdcSpillHeader
  {
    uint32_t fragmentSize;
    uint32_t fragmentType;
    uint32_t spillWordCount;
    uint8_t controllerNumber;
    uint8_t tdcNumber;
    uint32_t spillTriggerCounter;
    uint32_t spillStatus;
  } TdcSpillHeader_t;

  struct TdcEventHeader
  {
    uint32_t fragmentSize;
    uint32_t fragmentType;
    uint8_t  eventWordCount;
    uint8_t  eventStatus;
    uint32_t triggerCounter;
    uint8_t  triggerType;
    uint16_t controllerTimeStamp;
    uint32_t tdcTimeStamp;
    uint32_t nHits;
    uint8_t controllerNumber;
    uint8_t tdcNumber;
  } TdcEventHeader_t;

  struct TdcHit
  {
    uint8_t channel;
    uint16_t timeBin;
  } TdcHit_t;

#pragma pack(pop)

  void print(std::ostream & os = std::cout);

  static void printTdcControllerHeader(TdcControllerHeader header,
				       std::ostream & os = std::cout);
  static void printTdcSpillHeader(TdcSpillHeader header,
				  std::ostream & os = std::cout);
  static void printTdcEventHeader(TdcEventHeader header,
				  std::ostream & os = std::cout);
  static void printTdcHit(TdcHit hit, uint32_t index, 
			  std::ostream & os = std::cout);

  // Data from one TDC for one event
  class TdcEventData
  {
  public:
    struct TdcEventHeader tdcEventHeader;
    // Vector of Hits within a single trigger/event and this TDC
    std::vector<struct TDCFragment::TdcHit> tdcHits;  
  };

  // Data members of the TDCFragment class

  // One and only one controller header per spill:
  struct TdcControllerHeader controllerHeader;

  // Vector of up to 16 TDCs spill headers attached to controller
  // One spill header per TDC per spill
  std::vector<struct TdcSpillHeader> tdcSpillHeaders; 

  // Nested vector of TDC events; the top level vector is over events/triggers
  // The lower level vector is over TDCs within an event
  std::vector<std::vector<TdcEventData>> tdcEvents; 

  uint32_t streamToBuffer(std::ostream& buffer);
  uint32_t fillFromBuffer(std::istream& buffer);

};

#endif
