////////////////////////////////////////////////////////////////////////
// Class:       CAENADCDump
// Module Type: analyzer
// File:        CAENADCDump_module.cc
// Description: Prints out information about each event.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"

#include "art/Utilities/Exception.h"
#include "lariat-artdaq/Overlays/CAENADCFragment.hh"
#include "ModuleUtils.hh"
#include "artdaq-core/Data/Fragments.hh"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <fstream>
#include <iomanip>
#include <vector>
#include <iostream>

namespace lariat {
  class CAENADCDump;
}

class lariat::CAENADCDump : public art::EDAnalyzer {
public:
  explicit CAENADCDump(fhicl::ParameterSet const & pset);
  virtual ~CAENADCDump() {}

  virtual void analyze(art::Event const & evt);

private:
  std::string raw_data_label_;
  uint32_t num_adcs_to_show_;
};


lariat::CAENADCDump::CAENADCDump(fhicl::ParameterSet const & pset)
    : EDAnalyzer(pset),
      raw_data_label_(pset.get<std::string>("raw_data_label")),
      num_adcs_to_show_(pset.get<uint32_t>("num_adcs_to_show", 0))
		      
{
}

void lariat::CAENADCDump::analyze(art::Event const & evt)
{

  art::EventNumber_t spillNumber = evt.event();

  art::Handle<artdaq::Fragments> raw;
  evt.getByLabel(raw_data_label_, "SPILL", raw);

  if ( raw.isValid() ) {

    if (raw->size() != 1 ) {
      throw cet::exception("Error in CAENADCDump module: expected one and only one spill fragment");
    }
    
    const auto& frag((*raw)[0]);

    std::vector<const uint8_t*> lariatFragPtrs;

    LariatFragment::LariatFragmentType lariatFragType = LariatFragment::LariatFragmentType::FRAGMENT_TYPE_CAEN_ADC;

    getLariatFragments(frag, 
     		       lariatFragType,
		       lariatFragPtrs
		       );

    std::cout << std::endl;
    std::cout << "Run " << evt.run() << ", subrun " << evt.subRun()
       	      << ", spill " << spillNumber << " has " << lariatFragPtrs.size()
   	      << " fragment(s) of type " << lariatFragType << std::endl;
      
    for (size_t idx = 0; idx < lariatFragPtrs.size(); ++idx) {

      lariat::CAENADCFragment bb( lariatFragPtrs[idx]  );

      std::cout << std::endl;
      std::cout << "CAEN fragment " << idx << " has total ADC counts = " 
		<< bb.total_values() << std::endl;
      std::cout << std::endl;

      bb.printHeader();

      if (num_adcs_to_show_ > 0) {

	if (num_adcs_to_show_ > bb.total_values() ) {
	  throw cet::exception("num_adcs_to_show is larger than total number of adcs in fragment");
	} else {
	  std::cout << std::endl;
	  std::cout << "First " << num_adcs_to_show_ 
   		    << " ADC values in the fragment: " 
   		    << std::endl;
   	}

	for (uint32_t i_adc = 0; i_adc < num_adcs_to_show_; ++i_adc) {
	  std::cout << *(bb.dataBegin() + i_adc) << " ";
	}
      }

      std::cout << std::endl;
      std::cout << "######################################################################" << std::endl;
      std::cout << std::endl;
      std::cout << std::endl;
    }
  } else {
    throw cet::exception("Error in CAENADCDump_module: expected one and only one spill fragment");
  }

  std::cout << std::endl;

}

DEFINE_ART_MODULE(lariat::CAENADCDump)
