////////////////////////////////////////////////////////////////////////
// Class:       TDCDump
// Module Type: analyzer
// File:        TDCDump_module.cc
// Description: Prints out information about each event.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"

#include "art/Utilities/Exception.h"
#include "lariat-artdaq/Overlays/TDCEventFragment.hh"
#include "lariat-artdaq/Overlays/TDCSpillFragment.hh"
#include "lariat-artdaq/Overlays/TDCControllerFragment.hh"
#include "artdaq-core/Data/Fragments.hh"
#include "include/TDCFragment.h"
#include "include/LariatFragment.h"
#include "ModuleUtils.hh"

#include <algorithm>
#include <cassert>
#include <cmath>
#include <fstream>
#include <iomanip>
#include <vector>
#include <iostream>

namespace lariat {
  class TDCDump;
}

class lariat::TDCDump : public art::EDAnalyzer {
public:
  explicit TDCDump(fhicl::ParameterSet const & pset);
  virtual ~TDCDump() {}

  virtual void analyze(art::Event const & evt);

private:
  std::string raw_data_label_;
  uint32_t max_tdcs_to_show_;
};


lariat::TDCDump::TDCDump(fhicl::ParameterSet const & pset)
    : EDAnalyzer(pset),
      raw_data_label_(pset.get<std::string>("raw_data_label")),
      max_tdcs_to_show_(pset.get<uint32_t>("max_tdcs_to_show", 0))
		      
{
}

void lariat::TDCDump::analyze(art::Event const & evt)
{

  art::EventNumber_t spillNumber = evt.event();

  art::Handle<artdaq::Fragments> raw;
  evt.getByLabel(raw_data_label_, "SPILL", raw);

  if ( ! raw.isValid() || raw->size() != 1 ) {
    throw cet::exception("Error in TDCDump module: expected one and only one spill fragment");
  } else {
    
    const auto& spillfrag((*raw)[0]);
    
    std::vector<LariatFragment::LariatFragmentType> lariatFragTypes;

    lariatFragTypes.push_back( LariatFragment::LariatFragmentType::FRAGMENT_TYPE_TDC_CONTROLLER_HEADER );
    lariatFragTypes.push_back( LariatFragment::LariatFragmentType::FRAGMENT_TYPE_TDC_TDC_SPILL_HEADER );
    lariatFragTypes.push_back( LariatFragment::LariatFragmentType::FRAGMENT_TYPE_TDC_TDC_EVENT_HEADER );

    for ( auto lariatFragType : lariatFragTypes ) {

      std::vector<const uint8_t*> lariatFragPtrs;

      getLariatFragments(spillfrag, 
			 lariatFragType,
			 lariatFragPtrs
			 );

      std::cout << std::endl;
      std::cout << "Run " << evt.run() << ", subrun " << evt.subRun()
		<< ", spill " << spillNumber << " has " << lariatFragPtrs.size()
		<< " fragment(s) of type " << lariatFragType << std::endl;
      
      for (size_t idx = 0; idx < lariatFragPtrs.size(); ++idx) {

	switch ( lariatFragType ) {
	case LariatFragment::LariatFragmentType::FRAGMENT_TYPE_TDC_CONTROLLER_HEADER:
	  {
	    lariat::TDCControllerFragment tdcc( lariatFragPtrs[idx]  );

	    tdcc.printHeader();
	    break;
	  }
	case LariatFragment::LariatFragmentType::FRAGMENT_TYPE_TDC_TDC_SPILL_HEADER:
	  {
	    lariat::TDCSpillFragment tdcs( lariatFragPtrs[idx]  );

	    tdcs.printHeader();
	    break;
	  }
	case LariatFragment::LariatFragmentType::FRAGMENT_TYPE_TDC_TDC_EVENT_HEADER:
	  {
	    lariat::TDCEventFragment tdce( lariatFragPtrs[idx]  );

	    tdce.printHeader();

	    if (max_tdcs_to_show_ > 0) {

	      uint32_t actual_to_show = max_tdcs_to_show_ <= tdce.total_values() ?
		 	  max_tdcs_to_show_ : tdce.total_values() ;

	      std::cout << std::endl;
	      std::cout << "First " << actual_to_show
			<< " TDC values in the fragment: " 
			<< std::endl;

	      for (uint32_t i_tdc = 0; i_tdc < actual_to_show; ++i_tdc) {
		std::cout << "(Channel " << static_cast<int>( (*(tdce.dataBegin() + i_tdc)).channel ) << 
		  ", Timebin " << (*(tdce.dataBegin() + i_tdc)).timeBin << ") ";
	      }

	      std::cout << std::endl;
	      std::cout << "######################################################################" << std::endl;
	      std::cout << std::endl;
	      std::cout << std::endl;
	    }
	    break;
	  }
	default:
	  throw cet::exception( "Error in TDCDump: unknown fragment type requested" );
	}
      }

      std::cout << std::endl;

    }
  }
}

DEFINE_ART_MODULE(lariat::TDCDump)
