
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Run.h"
#include "art/Utilities/InputTag.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Services/Optional/TFileService.h"

#include "artdaq-core/Data/Fragments.hh"

#include "lariat-artdaq/Overlays/FragmentType.hh"
#include "lariat-artdaq/Overlays/CAENADCFragment.hh"
#include "lariat-artdaq/Overlays/GenericFragment.hh"

#include "lariat-artdaq/ArtModules/ModuleUtils.hh"

#include "include/LariatFragment.h"

#include "cetlib/exception.h"

#include "TRootCanvas.h"
#include "TCanvas.h"
#include "TGraph.h"
#include "TAxis.h"
#include "TH1D.h"
#include "TH2D.h"
#include "TStyle.h"
#include "TFile.h"
#include "TROOT.h"
#include "TStopwatch.h"

#include <boost/range/irange.hpp>

#include <numeric>
#include <vector>
#include <functional>
#include <algorithm>
#include <iostream>
#include <sstream>
#include <initializer_list>
#include <limits>
#include <unordered_set>
#include <set>

using std::cout;
using std::cerr;
using std::endl;

namespace lariat {

  class WFViewer : public art::EDAnalyzer {

  public:
    explicit WFViewer (fhicl::ParameterSet const & p);
    virtual ~WFViewer () = default;

    void analyze (art::Event const & e) override;
    void beginRun(art::Run const &) override;

  private:

    void CreateCanvases(std::size_t nplots);

    std::unique_ptr<TCanvas> canvas_[2];
    std::vector<Double_t> x_;
    int prescale_;
    bool digital_sum_only_;
    art::RunNumber_t current_run_;
    std::size_t nSamples_;
    std::size_t nChannels_;

    std::vector<artdaq::Fragment::fragment_id_t> fragment_ids_;

    std::string histos_filename_;
    std::string graphics_dirname_;

    bool use_timing_;
    bool live_paint_;
    
    //    art::ServiceHandle<art::TFileService> tfs_;
    std::vector<std::shared_ptr<TGraph>> graphs_;
    std::vector<std::shared_ptr<TH1D>> histograms_;
    std::vector<std::shared_ptr<TH2D>> twoDhistograms_;

    std::map<artdaq::Fragment::fragment_id_t, std::size_t> id_to_index_;

    std::size_t totalEventsSeen_;
    std::unordered_set<uint32_t> eventsSeenInSpill_;

  };

}

lariat::WFViewer::WFViewer (fhicl::ParameterSet const & ps): 
  art::EDAnalyzer(ps), 
  prescale_(ps.get<int> ("prescale")), 
  digital_sum_only_(ps.get<bool> ("digital_sum_only", false)), 
  current_run_(0), 
  nSamples_(ps.get<std::size_t>("num_samples")),
  nChannels_(64),
  histos_filename_(ps.get<std::string>("histos_filename","renameme_histos.root")),
  graphics_dirname_(ps.get<std::string>("graphics_dirname", ".")),
  use_timing_(static_cast<bool>( ps.get<int>("use_timing", 0)) ),
  live_paint_(static_cast<bool>( ps.get<int>("live_paint", 0)) ),
  totalEventsSeen_(0)
{

  gStyle->SetOptStat("irm");
  gStyle->SetMarkerStyle(22);
  gStyle->SetMarkerColor(4);

  if ( ! live_paint_ ) {
     gROOT->SetBatch(); // Don't actually paint the canvases, just write them to file
  }
}


void lariat::WFViewer::analyze (art::Event const & evt) {

  TStopwatch watch;
  watch.Start(true);

  art::EventNumber_t spillNumber = evt.event();


  std::size_t max_adc_count = pow(2, 12); // 12 bits for a V1740 
  std::size_t total_adc_values = nSamples_ * nChannels_;

  art::Handle<artdaq::Fragments> raw;

  evt.getByLabel("daq", "SPILL", raw);

  if ( !raw.isValid() || raw->size() != 1) {

    throw cet::exception("Error in WFViewer_module: expected one and only one spill fragment");

  } else {


    TFile f(histos_filename_.c_str(), "UPDATE");
    f.cd();

    eventsSeenInSpill_.clear();
    const auto& spillfrag((*raw)[0]);

    std::vector<const uint8_t*> lariatFragPtrs;

    // JCF, 7/7/14

    // First, grab the run number

    auto lariatFragType = 
      LariatFragment::LariatFragmentType::FRAGMENT_TYPE_RUNINFO ;

    getLariatFragments(spillfrag, lariatFragType, lariatFragPtrs);

    if (lariatFragPtrs.size() != 1) {
      std::cerr << "Found " << lariatFragPtrs.size() << " Run fragments" << std::endl;
      throw cet::exception("Error in WFViewer: expected one and only one Run fragment");
    }

    GenericFragment<LariatFragment::RunInfo> runInfo( lariatFragPtrs[0] );

    auto runNumber = runInfo.getHeader()->runNumber;

    lariatFragPtrs.clear();

    // Now get the V1740 fragments

    lariatFragType = LariatFragment::LariatFragmentType::FRAGMENT_TYPE_CAEN_ADC ;

    getLariatFragments(spillfrag,
                       lariatFragType,
                       lariatFragPtrs
                       );

    std::cout << std::endl;
    std::cout << "Run " << runNumber << ", subrun " << evt.subRun()
              << ", spill " << spillNumber << " has " << lariatFragPtrs.size()
              << " fragment(s) of type " << lariatFragType << std::endl;

    if (lariatFragPtrs.size() == 0) {
      f.Close();
      return;
    }


    // JCF, 7/7/14
    
    // If this is the first event we've seen, fill the fragment_ids_
    // vector with the board IDs of the V1740 boards, and create the
    // canvases which will contain plots of the V1740 data

    if (totalEventsSeen_ == 0) {

      // JCF, 6/23/14

      // Deduce the # of V1740 boards in use from the V1740 fragments

      std::set<uint32_t> boardIds;

      for (auto fragPtr : lariatFragPtrs) {

	const CAENADCFragment* caenPtr = new CAENADCFragment(reinterpret_cast<const uint8_t*>(fragPtr ));
	boardIds.insert( caenPtr->getHeader()->boardId );
      }

      std::cout << "WFViewer: There appear to be " << boardIds.size() << " V1740 boards" << std::endl;

      for (const auto& boardId : boardIds ) {
	fragment_ids_.emplace_back( static_cast<artdaq::Fragment::fragment_id_t>(boardId ));
      }

      CreateCanvases( fragment_ids_.size() );

      // id_to_index_ will translate between a fragment's ID and where in
      // the vector of graphs and histograms it's located

      for (std::size_t i_f = 0; i_f < fragment_ids_.size(); ++i_f) {
	id_to_index_[ fragment_ids_[i_f] ] = i_f;
      }

      graphs_.resize( fragment_ids_.size() );
      histograms_.resize( fragment_ids_.size() );
      twoDhistograms_.resize( fragment_ids_.size() );

      for (auto &x: graphs_) x = 0;
      for (auto &x: histograms_) x = 0;
      for (auto &x: twoDhistograms_) x = 0;
      
    } 


    if ( lariatFragPtrs.size() % fragment_ids_.size() != 0) {
      throw cet::exception("# of CAENADC fragments found is not a multiple of the number of fragment_ids supplied");
    } else {
      std::cout << "There appear to be " << lariatFragPtrs.size() / fragment_ids_.size() <<
	" triggered events in the spill" << std::endl;
    }


    // Plot the integrated distribution of ADC counts for each
    // fragment_id. Also, for every Nth triggered event, where N is
    // the "prescale" setting, if "digital_sum_only" is set to false
    // in the FHiCL string, then plot, for the Nth event, a graph of
    // the ADC values across all channels in each board_id /
    // fragment_id combo

    uint32_t eventCounter = std::numeric_limits<uint32_t>::max();
    uint32_t fragCounter = 0;
    
    for (auto fragPtr : lariatFragPtrs) {
    
      fragCounter++;

      const CAENADCFragment* caenPtr = new CAENADCFragment(reinterpret_cast<const uint8_t*>(fragPtr ));

      eventCounter = caenPtr->getHeader()->eventCounter;
      auto got = eventsSeenInSpill_.find( eventCounter );

      if ( got == eventsSeenInSpill_.end() ) {
	eventsSeenInSpill_.emplace( eventCounter );
	totalEventsSeen_++;
	//	std::cout << "totalEventSeen now " << totalEventsSeen_ << ", spill = " << spillNumber << ", event = " << eventCounter << std::endl;
      }


      artdaq::Fragment::fragment_id_t fragment_id = caenPtr->getHeader()->boardId;
      std::size_t ind = id_to_index_[ fragment_id ];

    
      // If a histogram doesn't exist for this fragment_id , create it

      if (!histograms_[ind]) {

	// histograms_[ind] = std::shared_ptr<TH1D>( 
	// 					 tfs_->make<TH1D>( Form ("Fragment_%d_hist", fragment_id), 
	// 							   "", max_adc_count+1, -0.5, 
	// 							   max_adc_count + 0.5 ) );

	histograms_[ind].reset( new TH1D(Form ("Fragment_%d_hist", fragment_id), 
					 "", max_adc_count+1, -0.5, 
					 max_adc_count + 0.5 ) );
	histograms_[ind]->SetTitle (Form ("Frag %d, Type CAENADC", fragment_id ) );
	histograms_[ind]->GetXaxis()->SetTitle("ADC value");
      }

      // For every event, fill the histogram (prescale is ignored here)
  
      for (auto val = caenPtr->dataBegin(); val != caenPtr->dataEnd(); ++val ) 
	histograms_[ind]->Fill( *val );

      if (totalEventsSeen_ % prescale_ - 1 != 0 && prescale_ > 1) {
	continue;
      }

      // If we pass the prescale, then if we're not going with
      // digital_sum_only, plot the ADC counts for this particular event/fragment_id
    
    if (!digital_sum_only_) {

      // Create the graph's x-axis

      if (x_.size () != total_adc_values) {
	x_.resize (total_adc_values);

	std::iota (x_.begin (), x_.end (), 0);
      }

      // If the graph doesn't exist, create it. Not sure whether to
      // make it an error if the total_adc_values is new

      if (!graphs_[ind] || static_cast<std::size_t>( graphs_[ind]->GetN () ) != total_adc_values) {
	//graphs_[ind] = std::shared_ptr<TGraph>( tfs_->make<TGraph>(total_adc_values) );
	graphs_[ind].reset( new TGraph(total_adc_values) );
	graphs_[ind]->SetName( Form ("Fragment_%d_graph", fragment_id));
	graphs_[ind]->SetLineColor ( 4 );
	std::copy (x_.begin (), x_.end (), graphs_[ind]->GetX ());
      }


      // Get the data from the fragment

      std::copy (caenPtr->dataBegin (), caenPtr->dataBegin() + total_adc_values, graphs_[ind]->GetY ());
      //      twoDhistograms_[ind] = std::shared_ptr<TH2D>( tfs_->make<TH2D>( Form("Fragment_%d_spill_%d_eventCounter_%d_twoDhist", fragment_id, spillNumber, eventCounter), "",
      //								      nSamples_, -0.5, 
      //								      nSamples_ - 0.5, nChannels_,
      //								      -0.5, nChannels_ - 0.5 ) );

      twoDhistograms_[ind].reset( new TH2D(Form("Fragment_%d_spill_%d_eventCounter_%d_twoDhist", fragment_id, spillNumber, eventCounter), "",
					   nSamples_, -0.5, 
					   nSamples_ - 0.5, nChannels_,
					   -0.5, nChannels_ - 0.5 ) );

      twoDhistograms_[ind]->SetTitle (Form ("Frag %d, Spill %d, Event %d", fragment_id, 
					    spillNumber, eventCounter) );
	
      twoDhistograms_[ind]->GetXaxis()->SetTitle( "Sample #" );
      twoDhistograms_[ind]->GetYaxis()->SetTitle( "Channel #" );

      twoDhistograms_[ind]->SetMinimum(0);
      twoDhistograms_[ind]->SetMaximum( max_adc_count );

      twoDhistograms_[ind]->SetStats(kFALSE);
      twoDhistograms_[ind]->Reset();

      if (caenPtr->total_values() != total_adc_values) {
	std::cerr << "Warning in WFViewer: expected " << nSamples_ << " samples but found " << caenPtr->total_values() / nChannels_ << std::endl;
      }

      for (size_t i_bin = 0; i_bin < caenPtr->total_values(); ++i_bin) {

	Int_t xbin = i_bin / nChannels_ + 1;
	Int_t ybin = i_bin % nChannels_ + 1;

	twoDhistograms_[ind]->SetBinContent( xbin, ybin, *( caenPtr->dataBegin() + i_bin) );
      }

      // And now prepare the graphics without actually drawing anything yet
      
      canvas_[1] -> cd(ind+1);
      TVirtualPad* pad = static_cast<TVirtualPad*>(canvas_[1]->GetPad(ind+1) ); 

      Double_t lo_x, hi_x, lo_y, hi_y, dummy;

      graphs_[ind]->GetPoint(0, lo_x, dummy);
      graphs_[ind]->GetPoint( graphs_[ind]->GetN()-1, hi_x, dummy);

      lo_x -= 0.5;
      hi_x += 0.5;

      lo_y = -0.5;
      hi_y = max_adc_count+0.5;


      TH1F* padframe = static_cast<TH1F*>( pad->DrawFrame( lo_x, lo_y, hi_x, hi_y ) );
      padframe->SetTitle( Form ("Frag %d, Spill %d, Event %d", static_cast<int>(fragment_id), 
				spillNumber, eventCounter ) );
      padframe->GetXaxis()->SetTitle("ADC #");
      pad->SetGrid();
      padframe->Draw("SAME");
    }

    // Draw the histogram

    canvas_[0]->cd(ind+1);
    histograms_[ind]->Draw();
    histograms_[ind]->Write();

    // And, if desired, the Nth event's ADC counts

    if (!digital_sum_only_) {
      canvas_[1]->cd(ind+1);

      twoDhistograms_[ind]->Draw("COLZ");
      twoDhistograms_[ind]->Write();
    }

    // JCF, 6/16/14

    // Only bother writing the canvas to file when all fragments are represented

    if (fragCounter % fragment_ids_.size() == 0) {

      std::stringstream filenamestr0;
      filenamestr0 << graphics_dirname_ << "/CAENV1740_R" << runNumber << "_sR" << evt.subRun() << ".pdf";
      canvas_[0]->Modified();
      canvas_[0]->Update();
      canvas_[0]->Print( filenamestr0.str().c_str() );


      if (!digital_sum_only_) {
	std::stringstream filenamestr1;
	filenamestr1 << graphics_dirname_ << "/CAENV1740_R" << runNumber << "_sR" << evt.subRun() << "_S" << spillNumber << "_E" << eventCounter << ".pdf";
	canvas_[1]->Modified();
	canvas_[1]->Update();
	canvas_[1]->Print( filenamestr1.str().c_str() );      
      }
    }
    }

    if (use_timing_) {
      watch.Stop();
      std::cout << "WFVIEWER MODULE TIME: " << std::endl;
      watch.Print();
    }


    f.Close();
  }

}

void lariat::WFViewer::beginRun(art::Run const &e) { 

  if (e.run () == current_run_) return;
  current_run_ = e.run ();

  for (int i = 0; i < 2; i++) canvas_[i] = 0;

}

void lariat::WFViewer::CreateCanvases(std::size_t nplots) {

  std::size_t num_x_plots = 0;
  std::size_t num_y_plots = 0;


  //  if (num_x_plots == std::numeric_limits<std::size_t>::max() ||
  //      num_y_plots == std::numeric_limits<std::size_t>::max() ) {

  switch ( nplots ) {
  case 1: num_x_plots = num_y_plots = 1; break;
  case 2: num_x_plots = 2; num_y_plots = 1; break;
  case 3:
  case 4: num_x_plots = 2; num_y_plots = 2; break;
  case 5: 
  case 6:  num_x_plots = 3; num_y_plots = 2; break;
  case 7:
  case 8: num_x_plots = 4; num_y_plots = 2; break;
  default: 
    num_x_plots = num_y_plots = static_cast<std::size_t>( ceil( sqrt( nplots ) ) );
  }
  //  }


  for (int i = 0; (i < 2 && !digital_sum_only_) || i < 1 ; i++) {
    canvas_[i] = std::unique_ptr<TCanvas>(new TCanvas(Form("wf%d",i)));
    canvas_[i]->Divide (num_x_plots, num_y_plots);
    canvas_[i]->Update ();
    ((TRootCanvas*)canvas_[i]->GetCanvasImp ())->DontCallClose ();
  }

  canvas_[0]->SetTitle("ADC Value Distribution");

  if (! digital_sum_only_) {
    canvas_[1]->SetTitle("ADC Values, Event Snapshot");
  }

}



// void countV1740s(const artdaq::Fragment& artdaqFrag) {

//   if (artdaqFrag.dataBegin() == artdaqFrag.dataEnd() ) {
//     throw cet::exception("Error in WFViewer:countV1740s: artdaq::Fragment appears to contain no payload");
//   }

//   // We'll need to use raw pointers rather than smart pointers,
//   // otherwise a crash will occur since other resources are being used to
//   // control the memory in the artdaq::Fragment

//   const uint8_t* bytePtr = 
//     reinterpret_cast<const uint8_t*>( &*artdaqFrag.dataBegin() );
  
//   const uint8_t* endPtr = reinterpret_cast< const uint8_t* >( &*artdaqFrag.dataEnd() );

//   // Subtract off the size of an artdaq::Fragment's dataType (8
//   // bytes), so that junk isn't read in past the end of the last
//   // lariat fragment...

//   while ( bytePtr < endPtr - sizeof(artdaq::RawDataType) ) {

//     const uint32_t* wordPtr = reinterpret_cast<const uint32_t*>( bytePtr );

//     uint32_t fragSize = *wordPtr;
//     uint32_t fragType = *(wordPtr+1);

//     if ( fragType == static_cast<uint32_t>( desiredFragType ) ) {
//       lariatPtrs.emplace_back( bytePtr );
//     }

//     bytePtr += fragSize;

//     //    std::cout << "Fragsize is " << fragSize << ", fragtype is " << fragType << ", bytePtr now at (decimal) " << reinterpret_cast<long>( bytePtr ) << std::endl;
//   }

// }




DEFINE_ART_MODULE(lariat::WFViewer)
