#include "V1495Fragment.h"
#include "LariatFragment.h"
#include <stdexcept>

V1495Fragment::V1495Fragment() {
  // give some sensible defaults so that
  // calling print on an unfilled V1495Fragment 
  // doesn't segv the program
  header.fragmentSize=0;
  header.fragmentType=LariatFragment::FRAGMENT_TYPE_V1495;
  header.nChan=0;
  header.nTrigPat=0;
  header.nVetoPat=0;
  header.M=2;
  header.tvMask=0xFFFF;
}

uint32_t V1495Fragment::streamToBuffer(std::ostream& buffer){

  uint32_t nchar_written=0;
  
  header.fragmentType = LariatFragment::FRAGMENT_TYPE_V1495;

  using std::streamsize;
  const streamsize headersize=sizeof(V1495Header_t);
  const streamsize channelsize = sizeof(V1495ChannelData_t);
  const streamsize patternsize=sizeof(V1495TriggerPatternData_t);
  header.fragmentSize = headersize 
    + channelsize*header.nChan
    + patternsize*header.nTrigPat
    + patternsize*header.nVetoPat;
  
  // write the header

  buffer.write(reinterpret_cast<const char*>(&header),headersize);
  nchar_written+=headersize;
  
  // write the channel data
  if(header.nChan!=channelData.size()) {
    throw std::length_error("V1495Fragment::streamToBuffer : header.nChan!=channelData.size()");
  }
  for(size_t i=0; i<header.nChan; ++i){
    V1495ChannelData_t data=channelData[i];
    buffer.write(reinterpret_cast<const char*>(&data),channelsize);
    nchar_written+=channelsize;
  }
  
  // write the trigger pattern data
    if(header.nTrigPat!=triggerPatternData.size()) {
    throw std::length_error("V1495Fragment::streamToBuffer : header.nTrigPat!=triggerPatternData.size()");
  }

  for(size_t i=0; i<header.nTrigPat; ++i){
    V1495TriggerPatternData_t data=triggerPatternData[i];
    buffer.write(reinterpret_cast<const char*>(&data),patternsize);
    nchar_written+=patternsize;
  }

  // write the veto pattern data
    if(header.nVetoPat!=vetoPatternData.size()) {
    throw std::length_error("V1495Fragment::streamToBuffer : header.nVetoPat!=vetoPatternData.size()");
  }

  for(size_t i=0; i<header.nVetoPat; ++i){
    V1495TriggerPatternData_t data=vetoPatternData[i];
    buffer.write(reinterpret_cast<const char*>(&data),patternsize);
    nchar_written+=patternsize;
  }
  return nchar_written;

}

uint32_t V1495Fragment::fillFromBuffer(std::istream& buffer) {

  uint32_t nchar_read=0;
  
  using std::streamsize;
  const streamsize headersize=sizeof(V1495Header_t);
  const streamsize channelsize = sizeof(V1495ChannelData_t);
  const streamsize patternsize=sizeof(V1495TriggerPatternData_t);
  
  //read the header

  buffer.read(reinterpret_cast<char*>(&header),headersize);
  nchar_read+=headersize;
  
  //read the channel data

  for(size_t i=0; i<header.nChan; ++i){
    V1495ChannelData_t data;
    buffer.read(reinterpret_cast<char*>(&data),channelsize);
    nchar_read+=channelsize;
    channelData.push_back(data);
  }
  if(header.nChan!=channelData.size()) {
    throw std::length_error("V1495Fragment::fillFromBuffer : header.nChan!=channelData.size()");
  }  

  //read the trigger pattern data
  for(size_t i=0; i<header.nTrigPat; ++i){
    V1495TriggerPatternData_t data;
    buffer.read(reinterpret_cast<char*>(&data),patternsize);
    nchar_read+=patternsize;
    triggerPatternData.push_back(data);
  }
  if(header.nTrigPat!=triggerPatternData.size()) {
    throw std::length_error("V1495Fragment::fillFromBuffer : header.nTrigPat!=triggerPatternData.size()");
  }

  // read the veto pattern data
  
  for(size_t i=0; i<header.nVetoPat; ++i){
    V1495TriggerPatternData_t data;
    buffer.read(reinterpret_cast<char*>(&data),patternsize);
    nchar_read+=patternsize;
    vetoPatternData.push_back(data);
  }
  if(header.nVetoPat!=vetoPatternData.size()) {
    throw std::length_error("V1495Fragment::fillFromBuffer : header.nVetoPat!=vetoPatternData.size()");
  }
  
  return nchar_read;
}

void V1495Fragment::print(std::ostream& os){
  using std::endl;
  using std::hex;
  using std::dec;

  os<<"=========== v1495Fragment ============="<<endl;
  os<<"fragmentSize : "<<header.fragmentSize<<endl;
  os<<"fragmentType : "<<header.fragmentType<<endl;
  os<<"nChan        : "<<header.nChan<<endl;
  os<<"nTrigPat     : "<<header.nTrigPat<<endl;
  os<<"nVetoPat     : "<<header.nVetoPat<<endl;
  os<<"M            : "<<header.M<<endl;
  os<<"tvMask       : 0x"<<hex<<header.tvMask<<dec<<endl;

  for(size_t i=0; i<header.nChan; ++i){
    os<<"*** Channel "<<channelData[i].num<<" ***"<<endl;
    os<<"Name   : "<<channelData[i].name<<endl;
    os<<"Counts : "<<channelData[i].counts<<endl;
  }
  
  for(size_t i=0; i<header.nTrigPat; ++i){
    os<<"*** Trigger "<<triggerPatternData[i].num<<" ***"<<endl;
    os<<"Pattern : 0x"<<hex<<triggerPatternData[i].pattern<<dec<<endl;
    os<<"Counts  : "<<triggerPatternData[i].counts<<endl;
  }

  for(size_t i=0; i<header.nVetoPat; ++i){
    os<<"*** Veto "<<vetoPatternData[i].num<<" ***"<<endl;
    os<<"Pattern : 0x"<<hex<<vetoPatternData[i].pattern<<dec<<endl;
    os<<"Counts  : "<<vetoPatternData[i].counts<<endl;
  }
  
}


void V1495Fragment::printHTML(int opt, std::ostream& os ){
  const std::string table_layout= "<table class=\"tg\"> \
  <tr> \
    <th class=\"tg-031e\">name1</th>\
    <th class=\"tg-031e\">name2</th>\
    <th class=\"tg-031e\">name3</th>\
    <th class=\"tg-031e\">name4</th>\
    <th class=\"tg-031e\">name5</th>\
    <th class=\"tg-031e\">name6</th>\
    <th class=\"tg-031e\">name7</th>\
    <th class=\"tg-031e\">name8</th>\
  </tr>\
  <tr>\
    <td class=\"tg-031e\">data1</td>\
    <td class=\"tg-031e\">data2</td>\
    <td class=\"tg-031e\">data3</td>\
    <td class=\"tg-031e\">data4</td>\
    <td class=\"tg-031e\">data5</td>\
    <td class=\"tg-031e\">data6</td>\
    <td class=\"tg-031e\">data7</td>\
    <td class=\"tg-031e\">data8</td>\
  </tr>\
  <tr>\
    <th class=\"tg-031e\">name9</td>\
    <th class=\"tg-031e\">name10</td>\
    <th class=\"tg-031e\">name11</td>\
    <th class=\"tg-031e\">name12</td>\
    <th class=\"tg-031e\">name13</td>\
    <th class=\"tg-031e\">name14</td>\
    <th class=\"tg-031e\">name15</td>\
    <th class=\"tg-031e\">name16</td>\
  </tr>\
  <tr>\
    <td class=\"tg-031e\">data9</td>\
    <td class=\"tg-031e\">data10</td>\
    <td class=\"tg-031e\">data11</td>\
    <td class=\"tg-031e\">data12</td>\
    <td class=\"tg-031e\">data13</td>\
    <td class=\"tg-031e\">data14</td>\
    <td class=\"tg-031e\">data15</td>\
    <td class=\"tg-031e\">data16</td>\
  </tr>\
</table>\
";


  std::string the_table=table_layout;
  std::vector<std::string> names;
  std::vector<std::string> datas;

  if(opt==0){
    // print the channel information
    for(size_t i=0; i<channelData.size(); ++i){
      names.push_back(std::string(channelData[i].name));
      char buf[256]; sprintf(buf,"%i",channelData[i].counts);
      datas.push_back(std::string(buf));
    }
  }
  else if(opt==1){
    for(size_t i=0; i<triggerPatternData.size(); ++i){
      char buf[256]; sprintf(buf,"Trigger %i",triggerPatternData[i].num);
      names.push_back(std::string(buf));
      char buf2[256]; sprintf(buf2,"%i",triggerPatternData[i].counts);
      datas.push_back(std::string(buf2));
    }
  }
  else if(opt==2){
    for(size_t i=0; i<vetoPatternData.size(); ++i){
      char buf[256]; sprintf(buf,"Veto %i",vetoPatternData[i].num);
      names.push_back(std::string(buf));
      char buf2[256]; sprintf(buf2,"%i",vetoPatternData[i].counts);
      datas.push_back(std::string(buf2));
    }
  }

  //  std::cout<<the_table<<std::endl;

  for(size_t i=0; i<names.size(); ++i){
    char buf[256]; sprintf(buf,">name%lu<",i+1);
    std::string name_to_replace(buf);
    char buf2[256]; sprintf(buf2,">data%lu<",i+1);
    std::string data_to_replace(buf2);
    
    //    std::cout<<"Looking for "<<name_to_replace<<std::endl;
    uint32_t p1=the_table.find(name_to_replace);
    std::string the_name = ">"; the_name+=names[i]; the_name+="<";
    the_table.replace(p1,name_to_replace.size(),the_name);

    //    std::cout<<"Looking for "<<data_to_replace<<std::endl;
    uint32_t p2=the_table.find(data_to_replace);
    std::string the_data = ">"; the_data+=datas[i]; the_data+="<";

    the_table.replace(p2,data_to_replace.size(),the_data);

  }

  os<<the_table<<std::endl;
}
